#include "circuit.h"
#include "simulator.h"
#include "paras.h"
#include "sat_atpg.h"
#include "checker.h"

using namespace atpg_ls;

int main(int argc, char *argv[]) {

    // 初始化命令行参数
    INIT_ARGS

    srand(OPT(seed));

    Circuit *circuit = new Circuit();

    Simulator* simulator = new Simulator();

    printf("parsing file %s ...\n", OPT(instance).c_str());

    simulator->parse_from_file(OPT(instance).c_str());
    simulator->insert_lines_for_stem();
    simulator->init_topo_index();
    
    circuit->parse_from_file(OPT(instance).c_str());
    circuit->insert_lines_for_stem();
    circuit->init_topo_index();

    circuit->simulator = simulator;
    circuit->init_faults();
    circuit->rebuild_atpg_circuit();

    init_final_checker();

    /**
     * D算法/电路 - 200000F = 时间优势好   (123) -> F -> (5,6)
     * SAT - 200000F * avg(N) (1011) (1010)
     * 
     * primary input 电路的输入
     * gate input 单个门的输入
     * 
     * (F0)
     * PIs assignment (10101010101010) -> (F0)
     * 
     * (10101010101010) inital sol
     * LOCAL + F0
     * 
     * (SAT + LS)
     **/
    sat_atpg_init(OPT(instance).c_str());
    
    // circuit->

    printf("building lut circuit ...\n");
    

    printf("====== Circuit Statistics ====== \n");
    printf("PIs:\t%ld\n", circuit->PIs.size());
    printf("POs:\t%ld\n", circuit->POs.size());
    printf("ori-Gate:\t%ld\n", circuit->original_gates.size());
    printf("new-Gate:\t%ld\n", circuit->gates.size());
    printf("================================ \n");

    circuit->ls_main();

    return 0;
}
